import sys,glob,re,math,os
from scipy.stats import describe

# This is the log of the total sample size (not only focal tree topology)
# Used in computing the posterior probability of the focal tree topology.
log_total = math.log(30000.)

# This is the log of the prior for tree topology. It is the log of 1 divided
# by the number of unrooted tree topologies for 6 taxa
#log_prob_T = -math.log(105.)

# The following are dictionaries each with key = model name and
# value = list of quantities, one from each replicate
logL = {} # log marginal likelihoods conditional on topology
logpT = {} # log posterior topology probability
logpY = {} # log marginal likelihood (unconditional)

class Table:
    def __init__(self):
        self.colname = []
        self.coldata = []
        self.nrows = 0
        self.ncols = 0
        
    def addColumn(self, name, vect):
        self.colname.append(name)
        self.coldata.append(vect)
        assert self.nrows == 0 or self.nrows == len(vect), 'all columns must have same length'
        if self.nrows == 0:
            self.nrows = len(vect)
        self.ncols += 1
        
    def show(self):
        assert self.nrows > 0, 'table has 0-length columns'

        # Print row of column headers
        print('%12s' % 'Model', end='')
        for col in range(self.ncols):
            print('\t%12s' % self.colname[col], end='')
        print('')

        # Print row of lines
        print('%12s' % '-----------', end='')
        for col in range(self.ncols):
            print('\t%12s' % '-----------', end='')
        print('')

        # Print body of table
        for row in range(self.nrows):
            print('%12d' % (row+1,), end='')
            for col in range(self.ncols):
                print('\t%12.5f' % self.coldata[col][row], end='')
            print('')
            
        # Print row of lines
        print('%12s' % '-----------', end='')
        for col in range(self.ncols):
            print('\t%12s' % '-----------', end='')
        print('')
        
        # Create vector of describe objects
        d = []
        for col in range(self.ncols):
            d.append(describe(self.coldata[col], 0, 1))
        
        # Output means for each column
        print('%12s' % "mean", end='')
        for col in range(self.ncols):
            mean = d[col].mean
            print('\t%12.5f' % mean, end='')
        print('')

        # Output standard deviations for each column
        print('%12s' % "sd", end='')
        for col in range(self.ncols):
            sd = math.sqrt(d[col].variance)
            print('\t%12.5f' % sd, end='')
        print('')

for i0 in range(30):
    i = i0 + 1
    subdir = 'g%d' % i
    filenames = glob.glob(os.path.join(subdir, 'loradML-*-output.txt'))
    print('subdir "%s"' % subdir)
    for fn in filenames:
        print('  ',fn)
        
        # Extract model name
        m = re.match('g\d+/loradML-(\S+)-output.txt', fn)
        model = 'unknown'
        if m is None:
            sys.exit('Error: could not extract model name from file name  "%s"' % fn)
        else:
            model = m.group(1)
        
        # Open file
        stuff = open(fn, 'r').read()

        # Extract log marginal likelihood (LoRaD)
        m = re.search('Log marginal likelihood is ([-.0-9]+)', stuff, re.M | re.S)
        if m is None:
            sys.exit('Error: could not extract log marginal likelihood from file name  "%s"' % fn)
        else:
            log_marg_like_given_T = float(m.group(1))
            if model in logL.keys():
                logL[model].append(log_marg_like_given_T)
            else:
                logL[model] = [log_marg_like_given_T]

        # Extract log posterior topology probability
        m = re.search('Sample size is (\d+)', stuff, re.M | re.S)
        if m is None:
            sys.exit('Error: could not extract sample size from file name  "%s"' % fn)
        else:
            sample_size = int(m.group(1))
            log_topol_post = math.log(sample_size) - log_total
            if model in logpT.keys():
                logpT[model].append(log_topol_post)
            else:
                logpT[model] = [log_topol_post]
                
        # Compute log marginal likelihood (unconditional)
        # Note: the formula on the line below may appear to be missing a term
        # (log_prob_T) but remember that LoRaD was applied to a sample file
        # from an MCMC analysis in which topology was allowed to vary (in which
        # all samples that did not include the focal topology were filtered out)
        # Thus, the quantity log_prob_T is already included in log_marg_like_given_T
        log_marg_like = log_marg_like_given_T - log_topol_post
        if model in logpY.keys():
            logpY[model].append(log_marg_like)
        else:
            logpY[model] = [log_marg_like]

model_names = sorted(logL.keys())
table = Table()
for m in model_names:
    table.addColumn(m, logpY[m])
table.show()
